#include <PalmOS.h>
#include <SysEvtMgr.h>
#include "PatNRsc.h"

#define secretY 60
#define lineY	(secretY + 16 + 4)
#define answerY (lineY + 4)
#define bottomY (answerY + 16 + 4)

#define ColumnsDefault 7
#define DurationDefault 120
#define durationNotRecorded 999
#define ArrayLength 12
#define ourMinVersion	sysMakeROMVersion(2,0,0,sysROMStageRelease,0)
#define checkpoint(x) WinDrawChars((x), 4, 20,20)
#define SEPARATABLE 1 /* enable */

typedef enum {
    NotStartedPhase = 0,
	PreIndicatePhase,
	IndicatePhase,
	PausePhase,
	AppearPhase,
	VanishPhase,
	AnswerPhase,
	ConfirmPhase,
	GameDonePhase,
	MaxPhase
} Phase;

UInt32 Wait[MaxPhase] = {
	evtWaitForever,     // NotStartedPhase
	80, 160, 160,		// by PreIndicate, Indicate, Pause,
	38, 10, //	  Appear, Vanish
	50, evtWaitForever, // Answer, Confirm
};

/* Internal Structures */
typedef struct {
	UInt16 columns;
	UInt16 duration;
} PatNPreferenceType;

typedef struct {
	UInt16 columns;
	UInt16 duration;
	UInt16 separation;
} RecordType;

typedef RecordType * RecordPtr;

/*
typedef struct {
	// Byte replaceme;
} PatNAppInfoType;

static PatNAppInfoType appInfo;
*/
static PatNPreferenceType prefs;

#define recordLength 23
RecordType record[recordLength] = {
	{7,durationNotRecorded,1},
	{8,durationNotRecorded,1},
	{6,durationNotRecorded,1},
	{9,durationNotRecorded,1},
	{5,durationNotRecorded,1},
	{4,durationNotRecorded,1},
	{3,durationNotRecorded,1},
	{2,durationNotRecorded,1},
	{1,durationNotRecorded,1},
	{6,durationNotRecorded,2},
	{6,durationNotRecorded,3},
	{6,durationNotRecorded,6},
	{7,durationNotRecorded,7},
	{9,durationNotRecorded,3},
	{8,durationNotRecorded,2},
	{8,durationNotRecorded,4},
	{4,durationNotRecorded,2},
	{4,durationNotRecorded,4},
	{2,durationNotRecorded,2},
	{3,durationNotRecorded,3},
	{5,durationNotRecorded,5},
	{8,durationNotRecorded,8},
	{9,durationNotRecorded,9},
};

short imageQTY;
Phase phase;
UInt32 waitTill;
Int32 duration;
UInt16 short columns;
UInt16 columnsEach;
UInt16 offset;
#define columnsQTY 9
UInt16 secret[columnsQTY];
UInt16 answer[columnsQTY];
UInt16 cursor;
Int16 incorrectCount;
UInt32 start;
UInt16 gameCount;
UInt16 gameQty = 10;
UInt16 separation = 1;

#define bitmapIDBase B0Bitmap
#define smallBitmapIDBase S0Bitmap
#define ButtonIDBase PatN0Button

BitmapPtr bitmapPtrs[ArrayLength];
MemHandle bitmapHandles[ArrayLength];
Char correctLabel[20], incorrectLabel[20], readyLabel[20];

/* Internal Constants */
#define DBName "PatN Score DB"
#define appFileCreator 'PatN'
#define appDbType 'SCOR'
#define appPrefID 0x00
#define appPrefVersionNum 0x01

// prototypes
static Err RomVersionCompatible(UInt32, UInt16);
static void* GetObjectPtr(UInt16);
static void indicateResult(UInt16);
static Boolean RoutineEvent(void);
static void MainFormInit(void);
static void MainFormFinalize(void);
static void indicateSettings(void);
static void getSettings(void);
static Boolean ButtonPressed(int);
static Boolean MainFormHandleEvent(EventPtr);
static Boolean HallOfFameFormHandleEvent(EventPtr);
static Boolean DrawHallOfFame();
static void clearScores(void);
static void readScores(void);
static void writeScores(void);
// static Boolean CreditFormHandleEvent(EventPtr);
static Boolean AppHandleEvent( EventType*);
static void AppEventLoop(void);
static Err AppStart(void);
static void AppStop(void);
UInt32 PilotMain(UInt16, MemPtr, UInt16);

/**
 * This routine checks that a ROM version is meet your minimum requirement.
 *
 * @param requiredVersion minimum rom version required
 *						  (see sysFtrNumROMVersion in SystemMgr.h
 *						  for format)
 * @param launchFlags	  flags that indicate if the application
 *						  UI is initialized.
 * @return error code or zero if rom is compatible
 */
static Err RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags) {
	UInt32 romVersion;
	// See if we're on in minimum required version of the ROM or later.
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion) {
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
				(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
			FrmAlert (SystemIncompatible);
			// Pilot 1.0 will continuously relaunch this app unless we switch to
			// another safe one.
			if (romVersion < sysMakeROMVersion(2,0,0,sysROMStageRelease,0)) {
				AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
			}
		}
		return sysErrRomIncompatible;
	}
	return 0;
}

/**
 * This routine returns a pointer to an object in the current form.
 * @param formId Id of the form to display
 * @return VoidPtr
 */
static void* GetObjectPtr(UInt16 objectID) {
	FormType* frmP;
	frmP = FrmGetActiveForm();
	return FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID));
}

/**
 * indicate whether correct or incorrect,
 * incorrect count in history.
 * @param incorrectCount the number of incorrect.
 */
static void indicateResult(UInt16 incorrectCount) {
	char buffer[2];
	Char* result;
	int i;
	incorrectCount %= 10; // round to one digit
	result = (incorrectCount==0)?correctLabel:incorrectLabel;
	WinDrawChars(result,StrLen(result),100,20);
	StrIToA(buffer, incorrectCount);
	WinDrawChars(buffer, 1, gameCount * 8 + 5, 120);
	if(incorrectCount == 0) {
		// record dureation
		for(i=0; i<recordLength; i++) {
			if(record[i].columns == columns &&
			   record[i].separation == separation &&
			   record[i].duration > duration) {
				  record[i].duration = duration;
				  break;
			}
		}
	}
}

/**
 * Routine Event for each phase
 */
static Boolean RoutineEvent() {
	Boolean handled = false;
	RectangleType bounds;
	FieldPtr fieldPtr;
	UInt32 answerDuration;
	char buffer[5];
	int i,xOffset, l,r;
	// UInt32 wait;
	switch (phase) {
	case PreIndicatePhase:
		// Get the settings from fields
		getSettings();
		// Generate the secret value
		for(i=0; i<columns; i++) {
			secret[i] = SysRandom(0) % (imageQTY);
		}
		incorrectCount = columns;
		// erase the place of labels
		bounds.topLeft.x = 80;
		bounds.topLeft.y = 20;
		bounds.extent.x = 160 - bounds.topLeft.x;
		bounds.extent.y = secretY - bounds.topLeft.y;
		WinEraseRectangle(&bounds,0);
		// erase the zone of secret values
		bounds.topLeft.x = 0;
		bounds.topLeft.y = secretY;
		bounds.extent.x = 160;
		bounds.extent.y = bottomY - secretY;
		WinEraseRectangle(&bounds,0);
		start = TimGetTicks();
		waitTill = start + Wait[phase];
		phase = IndicatePhase; // next
		handled = true;
		break;
	case IndicatePhase:
		WinDrawChars(readyLabel,StrLen(readyLabel),60,65);
		start = TimGetTicks();
		waitTill = start + duration; // Wait[phase];
		phase = PausePhase; // next
		handled = true;
		break;
	case PausePhase:
		if(TimGetTicks() < waitTill) break; // time not reached
		// hide the timing indicator
		bounds.topLeft.x = 0;
		bounds.topLeft.y = secretY;
		bounds.extent.x = 160;
		bounds.extent.y = 16;
		WinEraseRectangle(&bounds,0);
		WinEraseLine(0,lineY,160,lineY);
		// draw underlines
		xOffset = (160 - columns * 16) / 2;
		for(i=0; i < columns; i++) {
			l = xOffset + i * 16 + 1;
			r = xOffset + i * 16 + 14;
			WinDrawLine(l,lineY,r,lineY);
		}
		offset = 0;
		start = TimGetTicks();
		waitTill = start + duration; // Wait[phase];
		phase = AppearPhase; // next
		handled = true;
		break;
	case AppearPhase:
		if(TimGetTicks() < waitTill) break; // time not reached
		// indicate secret values
		bounds.topLeft.x = (160 - columnsEach * 16) / 2;
		bounds.topLeft.y = secretY;
		for(i=0; i<columnsEach; i++) {
			WinDrawBitmap(bitmapPtrs[secret[i + offset]],
				bounds.topLeft.x + i * 16, bounds.topLeft.y);
		}
		start = TimGetTicks();
		waitTill = start + duration / separation;// instead of Wait[phase];
		phase = VanishPhase; // next
		handled = true;
		break;
	case VanishPhase:
		// erase the secret values
		if(TimGetTicks() < waitTill) break; // time not reached
		bounds.topLeft.x = 0; // (160 - columnsEach * 16) / 2;
		bounds.topLeft.y = secretY;
		bounds.extent.x = 160; // 16 * columnsEach;
		bounds.extent.y = 16;
		WinEraseRectangle(&bounds, 0);
		offset += columns / separation;
		if(offset < columns) {
			waitTill = TimGetTicks() + duration / separation;
			phase = AppearPhase;
		} else {
			for(i=0; i<columns; i++) {
				answer[i] = -1;		// reset the answer
			}
			cursor = 0;
			start = TimGetTicks();
			waitTill = start + Wait[phase];
			phase = AnswerPhase; // next
		}
		handled = true;
		break;
	case AnswerPhase:
		// indicate duration of answering
		answerDuration = TimGetTicks() - start;
		if(answerDuration > 99999) answerDuration = 99999;
		StrIToA(buffer, answerDuration);
		WinDrawChars(buffer,
			answerDuration <	10 ? 1 :
			answerDuration <   100 ? 2 :
			answerDuration <  1000 ? 3 :
			answerDuration < 10000 ? 4 : 5
			,100,35);
		// if answer filled, go next
		if(cursor >= columns) {
			UInt32 wait;
			cursor=0;
			wait = 0; // answerDuration * 2 / 3 / columns ; //Wait[phase];
			// above line was modified because kubeyuri's suggestion...Thanx!
			if(wait > Wait[phase]) wait = Wait[phase]; // avoid Too long
			waitTill = TimGetTicks() + wait;
			phase = ConfirmPhase; // next
		} else {
			waitTill = TimGetTicks() + 12;
		}
		handled = true;
		break;
	case ConfirmPhase:
		if(TimGetTicks() < waitTill) break; // time not reached
		// indicate secret values again, for confirming
			waitTill = start + Wait[phase];
		bounds.topLeft.x = (160 - columns * 16) / 2 + cursor * 16;
		bounds.topLeft.y = secretY;
		WinDrawBitmap(bitmapPtrs[secret[cursor]],
			bounds.topLeft.x, bounds.topLeft.y);
		// Confirming
		if(answer[cursor] == secret[cursor]) {	// if correct
			incorrectCount -= 1;
		} else { // incorrect
			bounds.topLeft.y = answerY;
			WinDrawLine(
				bounds.topLeft.x, bounds.topLeft.y + 15,
				bounds.topLeft.x + 15, bounds.topLeft.y);
			WinDrawLine(
				bounds.topLeft.x, bounds.topLeft.y,
				bounds.topLeft.x + 15, bounds.topLeft.y + 15);
		}
		cursor++;
		if(cursor >= columns) {
			// finalize the game
			indicateResult(incorrectCount);
			waitTill = evtWaitForever;
			phase = GameDonePhase; // next
			duration += duration * (incorrectCount - 1) / 10;
			if(duration < 1) duration = 1;
			if(duration > 999) duration = 999;
			indicateSettings();
		}
		handled = true;
		break;
		/*
	case GameDonePhase:
		break;
	default:
		break;
		*/
   }
   return handled;
}

/**
 * get the values of column and duration
 */
static void getSettings() {
	FieldType* fieldPtr;
	MemHandle handle;
	Char* charPtr;
	ListType* listPtr;
	// get duration by duration field
	fieldPtr = (FieldType*)GetObjectPtr(PatNDurationField);
	handle = FldGetTextHandle(fieldPtr);
	if(handle != NULL) {
		charPtr = (Char*)MemHandleLock(handle);
		duration = StrAToI(charPtr);
		if(duration == 0) duration = DurationDefault;
		MemHandleUnlock(handle);
	}
	// get digits by digits list
	listPtr = (ListType*)GetObjectPtr(DigitsList);
	columns = LstGetSelection(listPtr) + 1; // zero origin
	if(columns == 0) columns = ColumnsDefault;
	if(separation > columns) separation = 1;
	columns = (columns / separation) * separation; // round to even value
	columnsEach = columns / separation;
	// re-indicate settings
	indicateSettings();
}
/**
 * indicate the values of column and duration.
 */
static void indicateSettings() {
	FieldType* fieldPtr;
	ControlType* popupPtr;
	ListType* listPtr;
	MemHandle handle;
	Char* charPtr;
	// draw the DurationField
	fieldPtr = (FieldType*)GetObjectPtr(PatNDurationField);
	FldFreeMemory(fieldPtr);
	handle = FldGetTextHandle(fieldPtr);
	if(handle == NULL) {
		handle = MemHandleNew(4);
	} else {
		MemHandleResize(handle,4);
	}
	charPtr = (Char*)MemHandleLock(handle);
	StrIToA(charPtr, duration);
	charPtr[((duration<10)?1:(duration<100)?2:3)] = '\0';
	MemHandleUnlock(handle);
	FldSetTextHandle(fieldPtr, handle);
	FldEraseField(fieldPtr);
	FldDrawField(fieldPtr);
	// update the List Cursor
	listPtr = (ListType*)GetObjectPtr(DigitsList);
	LstSetSelection(listPtr, columns-1); // zero origin
	// draw the ListPopup
	popupPtr = (ControlType*)GetObjectPtr(DigitsListPopup);
	charPtr = (Char*)CtlGetLabel(popupPtr);
	if (charPtr != NULL && *charPtr != '\0') {
		columns %= 10; // for safety
		*charPtr = chrDigitZero + columns;
		CtlDrawControl(popupPtr);
	}
}

/**
 * returns true if all columns are filled.
 */
static Boolean ButtonPressed(int number) {
	RectangleType bounds;
	answer[cursor] = number;
	bounds.topLeft.x = (160 - columns * 16) / 2;
	bounds.topLeft.y = answerY;
	WinDrawBitmap(bitmapPtrs[number],
		bounds.topLeft.x + cursor * 16, bounds.topLeft.y);
	return (++cursor >= columns);
}

/**
 * This routine is the event handler for the "PatNForm"
 * of this application.
 * @param eventP A pointer to an EventType structure
 * @return true if the event has handle and should not be passed
 *				to a higher level handler.
 */
static Boolean MainFormHandleEvent(EventType* eventP) {
	Boolean handled = false;
	FormType* frmP;
	UInt8 x,y;
	RectangleType bounds;
	ControlType* popupPtr;
	Char* charPtr;
	int newValue;
	int controlID;
	int i;
	FontID oldFont;
	MemHandle handle;
	BitmapPtr smallBitmapPtr;
	switch (eventP->eType)	{
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		FrmDrawForm ( frmP);
		bounds.topLeft.x =	100; bounds.topLeft.y =	20;
		bounds.extent.x = 40; bounds.extent.y = 11;
		WinEraseRectangle(&bounds,0);
		oldFont = FntSetFont(boldFont);
		for(i=0;i<imageQTY;i++) {
		    handle = DmGetResource(bitmapRsc,smallBitmapIDBase + i);
		    if(handle!=NULL) {
    		    smallBitmapPtr = (BitmapPtr)MemHandleLock(handle);
    			WinDrawBitmap(smallBitmapPtr,3+16*i,147);
    			MemHandleUnlock(handle);
        		DmReleaseResource(handle);
    		} else {
    		    WinDrawChar(chrDigitZero + i, 5+16*i, 147);
    		}
		}
		FntSetFont(oldFont);
		indicateSettings();
		// set the initial state
		waitTill = TimGetTicks() + Wait[PreIndicatePhase];
		phase = NotStartedPhase; // next
		handled |= RoutineEvent();
		break;
	case penDownEvent:
		x = eventP->screenX;
		y = eventP->screenY;
		if(phase == AnswerPhase && y >= secretY && y <= bottomY) {
			/* clear mistakes */
			newValue = ( x - (160 - columns * 16) / 2 ) / 16;
			if(newValue >= 0 && newValue < cursor) {
				cursor = newValue;
				bounds.topLeft.x =	(160 - columns * 16) / 2 + cursor * 16;
				bounds.topLeft.y =	answerY;
				bounds.extent.x = 160 - bounds.topLeft.x;
				bounds.extent.y = bottomY - answerY;
				WinEraseRectangle(&bounds,0);
				handled = true;
			}
		} else if (120 < y && y < 144 && x < 100) {
			/* tapped on history indicate zone ...
			   clear the history */
			gameCount = 0;
			bounds.topLeft.x = 0;
			bounds.extent.x = 96;
			bounds.topLeft.y = 120;
			bounds.extent.y = 12;
			WinEraseRectangle(&bounds,0);
			handled=true;
		} else if (phase > VanishPhase && SEPARATABLE && 80 < x && x < 140 && y < 20 ) {
			/* tapped on right top of the screen...
			   If Separatable; inclease separation */
			separation = (separation % 9) + 1;
			handled = true;
		}
		break;
	case ctlSelectEvent:
		controlID = eventP->data.ctlSelect.controlID;
		if(controlID == NewGameButton ) {
			// "New Game"
			if(phase > AppearPhase) {
				indicateResult(incorrectCount); // often re-indicate
				gameCount = ( gameCount + 1) % gameQty;
			}
			phase = PreIndicatePhase;
			handled |= RoutineEvent();
		} else if(controlID == AboutButton ) {
			// "About"
			FrmPopupForm(AboutForm);
			handled = true;
		} else if(controlID == DigitsListPopup) {
			// Number of Digits Selector
			newValue = LstPopupList(GetObjectPtr(DigitsList)) + 1;
			popupPtr = (ControlPtr)GetObjectPtr(DigitsListPopup);
			charPtr = (Char*)CtlGetLabel(popupPtr);
			if (charPtr != NULL && *charPtr != '\0') {
				*charPtr = chrDigitZero + newValue;
				CtlDrawControl(popupPtr);
			}
			handled = true;
		} else if(phase == AnswerPhase && ButtonIDBase <= controlID && controlID < ButtonIDBase + imageQTY ) {
			// number button is pressed
			SndPlaySystemSound(sndWarning);
			ButtonPressed(controlID - ButtonIDBase);
			handled |= RoutineEvent();
		} else if (phase == VanishPhase ||
				   phase == AppearPhase ||
				   phase == ConfirmPhase||
				   phase == PausePhase) {
			waitTill = TimGetTicks();
			handled |= RoutineEvent();
		}
		break;
	case frmCloseEvent:
	    MainFormFinalize();
	    break;
	case nilEvent:
		handled |= RoutineEvent();
		break;
		/*
	case keyDownEvent:
		if(phase == AnswerPhase) {
		}
		// PENDING: It well be handled, grafitti input
		break;
	default:
		break;
		*/
	}
	return handled;
}

/**
 * Blip the top-left corner; for Debugging
 */
 /*
static void Blip(int offset) {
	static Boolean blip = false;
	RectangleType rect;
	rect.topLeft.x = 10 * offset;
	rect.topLeft.y = 0;
	rect.extent.x = 10;
	rect.extent.y = 10;
	if(blip) {
		WinDrawRectangle(&rect,0);
	} else {
		WinEraseRectangle(&rect,0);
	}
	blip = !blip;
}
*/

/**
 * About Form
 */
static Boolean AboutFormHandleEvent(EventType* eventP) {
	Boolean handled = false;
	FormPtr frmP;
	switch (eventP->eType)	{
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		FrmDrawForm ( frmP);
		break;
	case ctlSelectEvent:
		switch (eventP->data.ctlSelect.controlID) {
			/*
		case CreditButton:
			FrmPopupForm(CreditForm);
			handled = true;
			break;
			*/
		case HallOfFameButton:
			FrmPopupForm(HallOfFameForm);
			handled = true;
			break;
		case CloseButton:
			FrmReturnToForm(MainForm);
			handled = true;
			break;
		default:
			break;
		}
		break;
	}
	return handled;
}

/**
 * Credit Form
 */
 /*
static Boolean CreditFormHandleEvent(EventType* eventP) {
	Boolean handled = false;
	FormPtr frmP;
	switch (eventP->eType)	{
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		FrmDrawForm ( frmP);
		break;
	case ctlSelectEvent:
		// I'm not check that what button is pressed.
		switch (eventP->data.ctlSelect.controlID) {
		case CloseButton:
			FrmReturnToForm (AboutForm);
			handled = true;
			break;
		default:
			break;
		}
		break;
	case frmCloseEvent:
		// do the default behaviour
		break;
	default:
		break;
	}
	return handled;
}
*/

/**
 * Hall Of Fame
 */
static Boolean HallOfFameFormHandleEvent(EventType* eventP) {
	Boolean handled = false;
	FormPtr frmP;
	UInt8 x,y;
	switch (eventP->eType)	{
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		FrmDrawForm ( frmP);
		DrawHallOfFame();
		handled = true;
		break;
	case ctlSelectEvent:
		// I'm not check that what button is pressed.
		switch (eventP->data.ctlSelect.controlID) {
		case CloseButton:
			FrmReturnToForm (AboutForm);
			handled = true;
			break;
		case ClearButton:
			clearScores();
			/*
			FrmDrawForm ( frmP);
			*/
			DrawHallOfFame();
			handled = true;
    		break;
			/*
		default:
			break;
		*/
		}
		break;
		/*
	case frmCloseEvent:
		// do the default behaviour
		break;
	default:
		break;
		*/
	}
	return handled;
}

/**
 * draw the duration records of specified separation
 */
static Boolean DrawHallOfFame() {
	int i;
	int columns;
	int duration;
	int separation;
	Char buffer[4];
	RectangleType rect;
	rect.topLeft.x = 20;
	rect.topLeft.y = 39;
	rect.extent.x = 140;
	rect.extent.y = 99;
	WinEraseRectangle(&rect, 1);
	for(i = 0 ; i < recordLength ; i++) {
		columns = record[i].columns;
		separation = record[i].separation;
		if(separation == 4 && columns == 8) separation = 3;
		if(separation > 3) separation = 4;
		duration = record[i].duration;
		if(duration != durationNotRecorded) {
			StrIToA(buffer, columns);
			WinDrawChars(buffer,1, 20, 28 + columns * 11);
			StrIToA(buffer, duration);
			WinDrawChars(buffer,duration <	  10 ? 1 :
								duration <	 100 ? 2 : 3
							,15 + 28 * separation , 28 + columns * 11);
		} else {
			WinDrawChars("---", 3
							,15 + 28 * separation , 28 + columns * 11);
		}
	}
}

static void clearScores() {
	LocalID dbID;
	int i;
	for(i = 0 ; i < recordLength ; i++) {
	    record[i].duration = durationNotRecorded;
	}
	dbID = DmFindDatabase(0,DBName);
	if(dbID != 0) DmDeleteDatabase(0,dbID);
}

/**
 * read scores from database
 */
static void readScores() {
	DmOpenRef recordsDB;
	UInt16 numRecords;
	RecordType* theRecord;
	MemHandle handle;
	int i,j;
	recordsDB = DmOpenDatabaseByTypeCreator(appDbType, appFileCreator, dmModeReadOnly);
	if(recordsDB == NULL) return;
	numRecords = DmNumRecords(recordsDB);
	for(i = 0 ; i < numRecords ; i++) {
		handle = DmQueryRecord(recordsDB, i);
		if(!handle) continue;
		theRecord = (RecordType*)MemHandleLock(handle);
		if(theRecord != NULL) {
			for(j = 0 ; j < recordLength; j++) {
				if((record[j].columns == theRecord->columns)
				 &&(record[j].separation == theRecord->separation)) {
					 record[j].duration = theRecord->duration;
					 break;
				}
			}
		MemHandleUnlock(handle);
		}
	}
	DmCloseDatabase(recordsDB);
}

/**
 * write scores to database
 */
static void writeScores() {
	DmOpenRef recordsDB;
	UInt16 numRecords;
	RecordType* theRecord;
	RecordType*	nilP = 0;
	MemHandle handle;
	UInt16 i,j;
	Err err;
	recordsDB = DmOpenDatabaseByTypeCreator(appDbType, appFileCreator, dmModeReadWrite);
	if(recordsDB != NULL) {
		// if the database exist, overwrite same record to it.
		numRecords = DmNumRecords(recordsDB);
		for(i = 0 ; i < numRecords ; i++) {
			handle = DmGetRecord(recordsDB, i);
			if(!handle) continue;
			theRecord = (RecordType*)MemHandleLock(handle);
			for(j = 0 ; j < recordLength; j++) {
				if((record[j].columns == theRecord->columns)
				 &&(record[j].separation == theRecord->separation)) {
					 DmWrite (theRecord, (UInt32)&nilP->duration, &record[j].duration, sizeof(nilP->duration));
					 record[j].duration = durationNotRecorded; // clear
					 break;
				}
			}
			MemHandleUnlock(handle);
			DmReleaseRecord(recordsDB, i, true);
		}
	} else {
		// or, if not exist, create new empty database.
		err = DmCreateDatabase (0, DBName, appFileCreator, appDbType, false);
		if (err) return;
		recordsDB = DmOpenDatabaseByTypeCreator(appDbType, appFileCreator, dmModeWrite);
		if (!recordsDB) return;
	}
	// then, write rest of the shortest record to the database.
	for(j = 0 ; j < recordLength; j++) {
		if(record[j].duration != durationNotRecorded) {
			handle = DmNewHandle(recordsDB, sizeof(RecordType));
			if (handle == NULL) continue;
			theRecord = MemHandleLock(handle);
			DmWrite (theRecord, (UInt32)&nilP->columns, &record[j].columns, sizeof(nilP->columns));
			DmWrite (theRecord, (UInt32)&nilP->separation, &record[j].separation, sizeof(nilP->separation));
			DmWrite (theRecord, (UInt32)&nilP->duration, &record[j].duration, sizeof(nilP->duration));
			MemHandleUnlock(handle);
			i = dmMaxRecordIndex;
			err = DmAttachRecord(recordsDB, &i, handle, NULL);
			if(err) MemHandleFree(handle);
		}
	}
  	DmCloseDatabase(recordsDB);
}

/**
 * This routine loads form resources and set the event handler for the form loaded.
 *
 * @param event A pointer to an EventType structure
 * @return true if the event has handle and should not be passed
 *				to a higher level handler.
 */
static Boolean AppHandleEvent( EventType* eventP) {
	UInt16 formId;
	FormPtr frmP;
	Boolean handled = false;
	switch (eventP->eType) {
	case frmLoadEvent:
		// Load the form resource.
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);
		switch (formId) {
		case MainForm:
			MainFormInit();
			FrmSetEventHandler(frmP, MainFormHandleEvent);
			handled = true;
			break;
		case AboutForm:
			FrmSetEventHandler(frmP, AboutFormHandleEvent);
			handled = true;
			break;
			/*
		case CreditForm:
			FrmSetEventHandler(frmP, CreditFormHandleEvent);
			handled = true;
			break;
			*/
		case HallOfFameForm:
			FrmSetEventHandler(frmP, HallOfFameFormHandleEvent);
			handled = true;
			break;
			/*
		default:
//			ErrFatalDisplay("Invalid Form Load Event");
			break;
			*/
		}
		break;
		/*
	default:
		break;
		*/
	}
	return handled;
}

/**
 * This routine is the event loop for the application.
 */
static void AppEventLoop(void) {
	UInt16 error;
	EventType event;
	UInt32 now, wait;
	do {
		now = TimGetTicks();
		wait = (waitTill == evtWaitForever) ? evtWaitForever :
			   (waitTill < now ) ? 0 : waitTill - now;
		EvtGetEvent(&event, wait);
		SysHandleEvent(&event) ||
//		MenuHandleEvent(0, &event, &error) ||
		AppHandleEvent(&event) ||
		FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

/**
 * Get the current application's preferences.
 * @return Err value 0 if nothing went wrong
 */
static Err AppStart(void) {
	MemHandle handle;
	UInt16 prefsSize;
	// intialize Game Status
	waitTill = evtWaitForever;
	gameCount = 0;
	// Read the saved preferences / saved-state information.
	prefsSize = sizeof(PatNPreferenceType);
	if (PrefGetAppPreferences(appFileCreator, appPrefID, &prefs, &prefsSize, true)
		 != noPreferenceFound) {
		duration = prefs.duration +
					((prefs.duration < DurationDefault) ? 20 : 0);
		if(duration > 999) duration = 999;
		columns = prefs.columns;
	} else {
		duration = DurationDefault;
		columns = ColumnsDefault;
	}
	if((handle = DmGetResource('tSTR',PatNCorrectLabel)) != NULL) {
		Char* consPtr = (Char*)MemHandleLock(handle);
		StrCopy(correctLabel,consPtr);
		MemHandleUnlock(handle);
		DmReleaseResource(handle);
	}
	if((handle = DmGetResource('tSTR',PatNIncorrectLabel)) != NULL) {
		Char* consPtr = (Char*)MemHandleLock(handle);
		StrCopy(incorrectLabel,consPtr);
		MemHandleUnlock(handle);
		DmReleaseResource(handle);
	}
	if((handle = DmGetResource('tSTR',PatNReadyLabel)) != NULL) {
		Char* consPtr = (Char*)MemHandleLock(handle);
		StrCopy(readyLabel,consPtr);
		MemHandleUnlock(handle);
		DmReleaseResource(handle);
	}
	// read records from the database
	readScores();
	return 0;
}

/**
 * Save the current state of the application.
 */
static void AppStop(void) {
	int i;
	// Write the saved preferences.
	prefs.duration = duration;
	prefs.columns = columns;
	PrefSetAppPreferences (appFileCreator, appPrefID, appPrefVersionNum,
		&prefs, sizeof (prefs), true);
	// write records to the database
	writeScores();
}

/**
 * This is the main entry point for the application.
 * @param cmd		  UInt16 value specifying the launch code.
 * @param cmdPB		  Pointer to a structure that is associated with the launch code.
 * @param launchFlags UInt16 value providing extra information about the launch.
 * @return Result of launch
 */
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags) {
	Err error;
	switch (cmd) {
	case sysAppLaunchCmdNormalLaunch:
		error = RomVersionCompatible (ourMinVersion, launchFlags);
		if (error) return error;
		error = AppStart();
		if (error) return error;
		FrmGotoForm(MainForm);
		AppEventLoop();
		AppStop();
		break;
	default:
		break;
	}
	return 0;
}


static void MainFormInit() {
	MemHandle handle;
	ControlType* buttonPtr;
	Char* buttonLabel;
	UInt32 depth;
	int i;
	Boolean enableColor;
	BitmapPtr smallBitmap;
	// get the image resource
	imageQTY = ArrayLength;
	// get ImageQTY from resource constant named 'tint'
	if((handle = DmGetResource('tint',ImageQTYID)) != NULL) {
		Char* consPtr = (Char*)MemHandleLock(handle);
		imageQTY = *consPtr;
		MemHandleUnlock(handle);
		DmReleaseResource(handle);
	}
	// Lock the bitmaps
	for(i=0; ( i < imageQTY) && (( bitmapHandles[i] = DmGetResource(bitmapRsc, bitmapIDBase + i)) != NULL) ;
			i++ ) {
				// PENDING: CAUTION! if you already lock the bitmapPtrs[i],
				// You must unlock MemPtr and Release the resource as following:
				//	 MemPtrUnlock(bitmapPtrs[i]);
				//	 DmReleaseResource(bitmapHandles[i]);
				// However I beleave it has not locked yet in this time.
		bitmapPtrs[i] = (BitmapPtr)MemHandleLock(bitmapHandles[i]);
	}
	imageQTY = i;
	if((handle = DmGetResource('tint',DepthID)) != NULL) {
		Char* consPtr = (Char*)MemHandleLock(handle);
		depth = *consPtr;
		MemHandleUnlock(handle);
		DmReleaseResource(handle);
		enableColor=true;
    	WinScreenMode(winScreenModeSet, NULL,NULL, &depth, &enableColor);
	}
}

static void MainFormFinalize() {
    int i;
	WinScreenMode(winScreenModeSetToDefaults, NULL,NULL,NULL,NULL);
	// Unlock the bitmaps
	for(i=0; bitmapHandles[i] != NULL ; i++) {
		MemPtrUnlock(bitmapPtrs[i]);
		DmReleaseResource(bitmapHandles[i]);
	}
}
